#include <pjsip.h>
#include <pjmedia.h>
#include <pjmedia-codec.h>
#include <pjsip_ua.h>
#include <pjsip_simple.h>
#include <pjlib-util.h>
#include <pjlib.h>
#include <pthread.h>

/* For logging purpose. */
#define THIS_FILE   "uac.c"
/* Settings */
#define AF	    pj_AF_INET()     /* Change to pj_AF_INET6() for IPv6.
				      										* PJ_HAS_IPV6 must be enabled and
				      										* your system must support IPv6.  */
#define SIP_PORT    5069	     /* Listening SIP port		*/
#define RTP_PORT    4009	     /* RTP port			*/
/*
 * Static variables.
 */

static pj_bool_t	     g_complete;    /* Quit flag.		*/
static pjsip_endpoint	    *g_endpt;	    /* SIP endpoint.		*/
static pj_caching_pool	     cp;	    /* Global pool factory.	*/

static pjmedia_endpt	    *g_med_endpt;   /* Media endpoint.		*/
static pjmedia_transport_info g_med_tpinfo; /* Socket info for media	*/
static pjmedia_transport    *g_med_transport;/* Media stream transport	*/
//static pj_bool_t	     g_complete;  

/* Call variables: */
static pjsip_inv_session    *g_inv;	    /* Current invite session.	*/
static pjmedia_session	    *g_med_session; /* Call's media session.	*/
/*forward declaration*/
static pj_bool_t on_rx_request( pjsip_rx_data *rdata );
static void on_tsx_state(pjsip_transaction *tsx, pjsip_event *e);
//IM global
pjsip_dialog *dlg;
pj_str_t IMto,IMfrom;


/* This is a PJSIP module to be registered by application to handle
 * incoming requests outside any dialogs/transactions. The main purpose
 * here is to handle incoming INVITE request message, where we will
 * create a dialog and INVITE session for it.
 */
static pjsip_module mod_sipim =
{
    NULL, NULL,			    /* prev, next.		*/
    { "mod-sipim", 9 },	    /* Name.			*/
    -1,				    /* Id			*/
    PJSIP_MOD_PRIORITY_APPLICATION, /* Priority			*/
    NULL,			    /* load()			*/
    NULL,			    /* start()			*/
    NULL,			    /* stop()			*/
    NULL,			    /* unload()			*/
    &on_rx_request,		    /* on_rx_request()		*/
    NULL,			    /* on_rx_response()		*/
    NULL,			    /* on_tx_request.		*/
    NULL,			    /* on_tx_response()		*/
    NULL,			    /* on_tsx_state()		*/
};



/* Util to display the error message for the specified error code  */
static int app_perror( const char *sender, const char *title, 
		       pj_status_t status)
{
    char errmsg[PJ_ERR_MSG_SIZE];

    pj_strerror(status, errmsg, sizeof(errmsg));

    PJ_LOG(3,(sender, "%s: %s [code=%d]", title, errmsg, status));
    return 1;
}
//we need to handle tsx state to get IM
static void on_tsx_state(pjsip_transaction *tsx, pjsip_event *e)
{
	printf("**************haha***************\n");
	//PJ_LOG(3,(dlg->obj_name, "------Transaction %s state changed to %s-----",tsx->obj_name, pjsip_tsx_state_str(tsx->state)));
}

/*
 * Callback when INVITE session state has changed.
 * This callback is registered when the invite session module is initialized.
 * We mostly want to know when the invite session has been disconnected,
 * so that we can quit the application.
 */
static void call_on_state_changed(pjsip_inv_session *inv,pjsip_event *e)
{
	PJ_UNUSED_ARG(e);

	if (inv->state == PJSIP_INV_STATE_DISCONNECTED) 
	{

		PJ_LOG(3,(THIS_FILE, "Call DISCONNECTED [reason=%d (%s)]",inv->cause,pjsip_get_status_text(inv->cause)->ptr));
		PJ_LOG(3,(THIS_FILE, "One call completed, application quitting..."));
		g_complete = 1;

	} 
	else 
	{
		PJ_LOG(3,(THIS_FILE, "Call state changed to %s", pjsip_inv_state_name(inv->state)));
  }
}

/* This callback is called when dialog has forked. */
static void call_on_forked(pjsip_inv_session *inv, pjsip_event *e)
{
	/* To be done... */
	PJ_UNUSED_ARG(inv);
	PJ_UNUSED_ARG(e);
}

/*
 * Callback when incoming requests outside any transactions and any
 * dialogs are received. We're only interested to hande incoming INVITE
 * request, and we'll reject any other requests with 500 response.
 */
static pj_bool_t on_rx_request( pjsip_rx_data *rdata )
{
	pj_sockaddr hostaddr;
	char temp[80], hostip[PJ_INET6_ADDRSTRLEN];
	pj_str_t local_uri;
	
	pjsip_tx_data *tdata;
	unsigned options = 0;
	pj_status_t status;
	pjmedia_sdp_session *local_sdp;


   
    /* 
     * Respond (statelessly) any non-INVITE requests with 500 
     */
  if (rdata->msg_info.msg->line.req.method.id == PJSIP_OTHER_METHOD) 
	{
		
		pjsip_msg_body *body = rdata->msg_info.msg->body;
		printf("recv message:%s\n",body->data); 
		pjsip_endpt_respond_stateless(g_endpt, rdata, 200, NULL,NULL, NULL);
		return PJ_TRUE;
	}
	if (rdata->msg_info.msg->line.req.method.id != PJSIP_INVITE_METHOD) 
	{

		if (rdata->msg_info.msg->line.req.method.id != PJSIP_ACK_METHOD) 
		{
	    pj_str_t reason = pj_str("Simple UA unable to handle this request");

	    pjsip_endpt_respond_stateless( g_endpt, rdata, 500, &reason,NULL, NULL);
		}
		return PJ_TRUE;
   }
    /*
     * Reject INVITE if we already have an INVITE session in progress.
     */
  if (g_inv) 
  {
		pj_str_t reason = pj_str("Another call is in progress");
		pjsip_endpt_respond_stateless( g_endpt, rdata,500, &reason,NULL, NULL);
		return PJ_TRUE;
   }

	/* Verify that we can handle the request. */
	status = pjsip_inv_verify_request(rdata, &options, NULL, NULL,g_endpt, NULL);
  if (status != PJ_SUCCESS) 
  {
		pj_str_t reason = pj_str("Sorry Simple UA can not handle this INVITE");
		pjsip_endpt_respond_stateless( g_endpt, rdata, 500, &reason,NULL, NULL);
		return PJ_TRUE;
   } 

    /*
     * Generate Contact URI
     */
	if (pj_gethostip(AF, &hostaddr) != PJ_SUCCESS) 
	{
		app_perror(THIS_FILE, "Unable to retrieve local host IP", status);
		return PJ_TRUE;
	}
  pj_sockaddr_print(&hostaddr, hostip, sizeof(hostip), 2);

  
  pj_ansi_sprintf(temp, "<sip:simpleuas@%s:%d>",hostip, SIP_PORT); 
  local_uri = pj_str(temp);
	
	
	pjsip_contact_hdr *contact_hdr ;
	contact_hdr = (pjsip_contact_hdr*)pjsip_msg_find_hdr(rdata->msg_info.msg, PJSIP_H_CONTACT,NULL);
  if (contact_hdr) 
  {
		IMto.ptr = (char*) pj_pool_alloc(rdata->tp_info.pool,PJSIP_MAX_URL_SIZE);
		IMto.slen = pjsip_uri_print(PJSIP_URI_IN_CONTACT_HDR,contact_hdr->uri, IMto.ptr,PJSIP_MAX_URL_SIZE);
   } 
   else 
   {
			printf("no contact information\n");
   }
   
    /*
     * Create UAS dialog.
     */
  status = pjsip_dlg_create_uas( pjsip_ua_instance(), rdata,&local_uri, /* contact */&dlg);
  if (status != PJ_SUCCESS) 
  {
  	pjsip_endpt_respond_stateless(g_endpt, rdata, 500, NULL,NULL, NULL);
		return PJ_TRUE;
	}
	
	/* 
     * Get media capability from media endpoint: 
     */

  status = pjmedia_endpt_create_sdp( g_med_endpt, rdata->tp_info.pool, 1,&g_med_tpinfo.sock_info, &local_sdp);
  PJ_ASSERT_RETURN(status == PJ_SUCCESS, PJ_TRUE);
    /* 
     * Create invite session, and pass both the UAS dialog and the SDP
     * capability to the session.
     */
	status = pjsip_inv_create_uas( dlg, rdata, local_sdp, 0, &g_inv);
	PJ_ASSERT_RETURN(status == PJ_SUCCESS, PJ_TRUE);


    /*
     * Initially send 180 response.
     *
     * The very first response to an INVITE must be created with
     * pjsip_inv_initial_answer(). Subsequent responses to the same
     * transaction MUST use pjsip_inv_answer().
     */
	status = pjsip_inv_initial_answer(g_inv, rdata,180, NULL, NULL, &tdata);
	PJ_ASSERT_RETURN(status == PJ_SUCCESS, PJ_TRUE);
  /* Send the 180 response. */  
  status = pjsip_inv_send_msg(g_inv, tdata); 
	PJ_ASSERT_RETURN(status == PJ_SUCCESS, PJ_TRUE);
    /*
     * Now create 200 response.
     */
	status = pjsip_inv_answer( g_inv, 200, NULL,/* st_code and st_text */NULL,/* SDP already specified */&tdata);
	PJ_ASSERT_RETURN(status == PJ_SUCCESS, PJ_TRUE);

    /*
     * Send the 200 response.
     */
	status = pjsip_inv_send_msg(g_inv, tdata);
	PJ_ASSERT_RETURN(status == PJ_SUCCESS, PJ_TRUE);

    /* Done. 
     * When the call is disconnected, it will be reported via the callback.
     */

	return PJ_TRUE;
}

static void call_on_media_update( pjsip_inv_session *inv,pj_status_t status)
{
	printf("************media update uac*******************\n");
}

pj_status_t SendIM(pjsip_dialog *dlg,pj_str_t text)
{
	pjsip_method MessageMethod={PJSIP_OTHER_METHOD,{"MESSAGE",7}};
	pjsip_tx_data *tdata;
	pj_status_t status=PJ_TRUE;
	char temp[80];
	
	
	status=pjsip_endpt_create_request(g_endpt,&MessageMethod,&IMto,&IMfrom,&IMto,&IMfrom,NULL,-1,&text,&tdata);
	if (status!=PJ_SUCCESS) 
		return status;
	status=pjsip_endpt_send_request(g_endpt,tdata,-1,NULL,NULL);
	return status;
}

void *thread_func(void *arg)
{
	char option[10];
	char msg[1024];
	pj_status_t status;
	
	pj_thread_t *mythread;
	static pj_thread_desc handler_desc;
	pj_thread_register("thread_func", handler_desc, &mythread);
	while (1)
	{
		puts("Press 'h' to hangup call");
		puts("Press 'm' then input message");
		fgets(option, sizeof(option), stdin);

		if (option[0] == 'h')
		{
			pjsip_tx_data *tdata;
			
			status = pjsip_inv_end_session(g_inv, PJSIP_SC_OK, NULL, &tdata);
			if (status == PJ_SUCCESS && tdata)
	  		status = pjsip_inv_send_msg(g_inv, tdata);
	  	//break;
		}
		if (option[0] == 'm')
		{
			fgets(msg,sizeof(msg),stdin);
			if (SendIM(dlg,pj_str(msg))!=PJ_SUCCESS)
				PJ_LOG(3,(THIS_FILE, "unable to send IM"));
		}
	}
}

int main(int argc, char *argv[])
{
	pj_status_t status;
	pj_sockaddr hostaddr;
	char hostip[PJ_INET6_ADDRSTRLEN+2];
	char temp[80];
	pj_str_t local_uri;
	
	/* Must init PJLIB first: */
	status = pj_init();
	PJ_ASSERT_RETURN(status == PJ_SUCCESS, 1);


	/* Then init PJLIB-UTIL: */
	status = pjlib_util_init();
	PJ_ASSERT_RETURN(status == PJ_SUCCESS, 1);
	/* Must create a pool factory before we can allocate any memory. */
	pj_caching_pool_init(&cp, &pj_pool_factory_default_policy, 0);
	/* Create global endpoint: */
  {
		const pj_str_t *hostname;
		const char *endpt_name;

	/* Endpoint MUST be assigned a globally unique name.
	 * The name will be used as the hostname in Warning header.
	 */

		/* For this implementation, we'll use hostname for simplicity */
		hostname = pj_gethostname();
		endpt_name = hostname->ptr;

		/* Create the endpoint: */

		status = pjsip_endpt_create(&cp.factory, endpt_name, &g_endpt);
		PJ_ASSERT_RETURN(status == PJ_SUCCESS, 1);
   }
    /* 
     * Add UDP transport, with hard-coded port 
     * Alternatively, application can use pjsip_udp_transport_attach() to
     * start UDP transport, if it already has an UDP socket (e.g. after it
     * resolves the address with STUN).
     */
 	{
		pj_sockaddr addr;

		pj_sockaddr_init(AF, &addr, NULL, (pj_uint16_t)SIP_PORT);
	
		if (AF == pj_AF_INET()) 
		{
	    status = pjsip_udp_transport_start( g_endpt, &addr.ipv4, NULL, 1, NULL);
		} 
		else if (AF == pj_AF_INET6()) 
		{
	    status = pjsip_udp_transport_start6(g_endpt, &addr.ipv6, NULL,1, NULL);
		} 
		else 
		{
	    status = PJ_EAFNOTSUP;
		}

		if (status != PJ_SUCCESS) 
		{
	  	app_perror(THIS_FILE, "Unable to start UDP transport", status);
	    return 1;
		}
	}


    /* 
     * Init transaction layer.
     * This will create/initialize transaction hash tables etc.
     */
  status = pjsip_tsx_layer_init_module(g_endpt);
	PJ_ASSERT_RETURN(status == PJ_SUCCESS, 1);


    /* 
     * Initialize UA layer module.
     * This will create/initialize dialog hash tables etc.
     */
  status = pjsip_ua_init_module( g_endpt, NULL );
  PJ_ASSERT_RETURN(status == PJ_SUCCESS, 1);


    /* 
     * Init invite session module.
     * The invite session module initialization takes additional argument,
     * i.e. a structure containing callbacks to be called on specific
     * occurence of events.
     *
     * The on_state_changed and on_new_session callbacks are mandatory.
     * Application must supply the callback function.
     *
     * We use on_media_update() callback in this application to start
     * media transmission.
     */
  {
		pjsip_inv_callback inv_cb;

		/* Init the callback for INVITE session: */
		pj_bzero(&inv_cb, sizeof(inv_cb));
		inv_cb.on_state_changed = &call_on_state_changed;
		inv_cb.on_new_session = &call_on_forked;
		inv_cb.on_media_update = &call_on_media_update;

		/* Initialize invite session module:  */
		status = pjsip_inv_usage_init(g_endpt, &inv_cb);
		PJ_ASSERT_RETURN(status == PJ_SUCCESS, 1);
   }

   /* Initialize 100rel support */
	status = pjsip_100rel_init_module(g_endpt);
	PJ_ASSERT_RETURN(status == PJ_SUCCESS, status);

    /*
     * Register our module to receive incoming requests.
     */
	status = pjsip_endpt_register_module( g_endpt, &mod_sipim);
	PJ_ASSERT_RETURN(status == PJ_SUCCESS, 1);
	
	#if PJ_HAS_THREADS
    status = pjmedia_endpt_create(&cp.factory, NULL, 1, &g_med_endpt);
	#else
    status = pjmedia_endpt_create(&cp.factory, 
				  pjsip_endpt_get_ioqueue(g_endpt), 
				  0, &g_med_endpt);
	#endif
    PJ_ASSERT_RETURN(status == PJ_SUCCESS, 1);
  
  /* 
  * Add PCMA/PCMU codec to the media endpoint. 
  */
     
  status = pjmedia_codec_g711_init(g_med_endpt);
  PJ_ASSERT_RETURN(status == PJ_SUCCESS, 1);
	
	/* 
     * Create media transport used to send/receive RTP/RTCP socket.
     * One media transport is needed for each call. Application may
     * opt to re-use the same media transport for subsequent calls.
     */
  status = pjmedia_transport_udp_create3(g_med_endpt, AF, NULL, NULL,RTP_PORT, 0, &g_med_transport);
  if (status != PJ_SUCCESS) 
  {
		app_perror(THIS_FILE, "Unable to create media transport", status);
		return 1;
   }

    /* 
     * Get socket info (address, port) of the media transport. We will
     * need this info to create SDP (i.e. the address and port info in
     * the SDP).
     */
   pjmedia_transport_info_init(&g_med_tpinfo);
   pjmedia_transport_get_info(g_med_transport, &g_med_tpinfo);

   if (pj_gethostip(AF, &hostaddr) != PJ_SUCCESS) 
		{
	    app_perror(THIS_FILE, "Unable to retrieve local host IP", status);
	    return 1;
		}
		pj_sockaddr_print(&hostaddr, hostip, sizeof(hostip), 2);

		pj_ansi_sprintf(temp, "<sip:simpleuac@%s:%d>", hostip, SIP_PORT);
		local_uri = pj_str(temp);
		IMfrom= pj_str(temp);
   
    /*
     * If URL is specified, then make call immediately.
     */
  if (argc > 1) 
  {
		
		pj_str_t dst_uri = pj_str(argv[1]);
		
		//pjsip_dialog *dlg;
		pjsip_tx_data *tdata;
		pjmedia_sdp_session *local_sdp;

		
		IMto=pj_str(argv[1]);
		/* Create UAC dialog */
		status = pjsip_dlg_create_uac( pjsip_ua_instance(), 
				       &local_uri,  /* local URI */
				       &local_uri,  /* local Contact */
				       &dst_uri,    /* remote URI */
				       &dst_uri,    /* remote target */
				       &dlg);	    /* dialog */
		if (status != PJ_SUCCESS) 
		{
	    app_perror(THIS_FILE, "Unable to create UAC dialog", status);
	    return 1;
		}
	
	/* Get the SDP body to be put in the outgoing INVITE, by asking
	 * media endpoint to create one for us. The SDP will contain all
	 * codecs that have been registered to it (in this case, only
	 * PCMA and PCMU), plus telephony event.
	 */
	status = pjmedia_endpt_create_sdp( g_med_endpt,	    /* the media endpt	*/
					   dlg->pool,	    /* pool.		*/
					   1,		    /* # of streams	*/
					   &g_med_tpinfo.sock_info,   
							    /* RTP sock info	*/
					   &local_sdp);	    /* the SDP result	*/
	PJ_ASSERT_RETURN(status == PJ_SUCCESS, 1);
	/* Create the INVITE session, and pass the SDP returned earlier
	 * as the session's initial capability.
	 */
		status = pjsip_inv_create_uac( dlg, local_sdp, 0, &g_inv);
		PJ_ASSERT_RETURN(status == PJ_SUCCESS, 1);

	

	/* Create initial INVITE request.
	 * This INVITE request will contain a perfectly good request and 
	 * an SDP body as well.
	 */
		status = pjsip_inv_invite(g_inv, &tdata);
		PJ_ASSERT_RETURN(status == PJ_SUCCESS, 1);



	/* Send initial INVITE request. 
	 * From now on, the invite session's state will be reported to us
	 * via the invite session callbacks.
	 */
		status = pjsip_inv_send_msg(g_inv, tdata);
		PJ_ASSERT_RETURN(status == PJ_SUCCESS, 1);
		} 
	else 
	{
			/* No URL to make call to */

		PJ_LOG(3,(THIS_FILE, "Ready to accept incoming calls..."));
	}

	pthread_t ta;
	
	pthread_create(&ta,NULL,thread_func,NULL);
	pthread_detach(ta);
	
	 
	/* Loop until one call is completed */
	for (;!g_complete;) 
	{
		
		pj_time_val timeout = {0, 10};
		pjsip_endpt_handle_events(g_endpt, &timeout);
		
		
	}
	return 0;
}


